#!/usr/bin/env python3

import os
import json
from enum import Enum

import importlib.util

class TimestampGranularity(Enum):
    MICROS = 1
    MILLIS = 2
    SECONDS = 3
    MINUTES = 4
    HOURS = 5
    HALF_DAYS = 6
    DAYS = 7

def get_timestamp_granularity(value):
    s = value.upper()
    ret = TimestampGranularity.MICROS
    if s == "MILLIS":
        ret = TimestampGranularity.MILLIS
    elif s == "SECONDS":
        ret = TimestampGranularity.SECONDS
    elif s == "MINUTES":
        ret = TimestampGranularity.MINUTES
    elif s == "HOURS":
        ret = TimestampGranularity.HOURS
    elif s == "HALF_DAYS":
        ret = TimestampGranularity.HALF_DAYS
    elif s == "DAYS":
        ret = TimestampGranularity.DAYS
    else:
        ret = TimestampGranularity.MICROS
    
    return ret

class configInfo:

    def __init__(self, event_reporting_wake_interval, event_reporting_minimum_count, event_reporting_flush_interval, 
                 event_reporting_trap_exceptions, event_reporting_timestamp_granularity, event_reporting_synchronous,
                 logging_verbose, key_caching_unstructured, key_caching_structured, key_caching_encrypt, 
                 key_caching_ttl_seconds, idp_provider, idp_ubiq_customer_id, idp_token_endpoint_url, idp_tenant_id,
                 idp_client_secret):
        self.__event_reporting_wake_interval = event_reporting_wake_interval
        self.__event_reporting_minimum_count = event_reporting_minimum_count
        self.__event_reporting_flush_interval = event_reporting_flush_interval
        self.__event_reporting_trap_exceptions = event_reporting_trap_exceptions
        self.__event_reporting_timestamp_granularity = event_reporting_timestamp_granularity
        self.__event_reporting_synchronous = event_reporting_synchronous
        self.__logging_verbose = logging_verbose
        self.__key_caching_unstructured = key_caching_unstructured
        self.__key_caching_structured = key_caching_structured
        self.__key_caching_encrypt = key_caching_encrypt
        self.__key_caching_ttl_seconds = key_caching_ttl_seconds
        self.__idp_provider = idp_provider
        self.__idp_ubiq_customer_id = idp_ubiq_customer_id
        self.__idp_token_endpoint_url = idp_token_endpoint_url
        self.__idp_tenant_id = idp_tenant_id
        self.__idp_client_secret = idp_client_secret

    def get_event_reporting_wake_interval(self):
        return self.__event_reporting_wake_interval
    event_reporting_wake_interval=property(get_event_reporting_wake_interval)

    def get_event_reporting_minimum_count(self):
        return self.__event_reporting_minimum_count
    event_reporting_minimum_count=property(get_event_reporting_minimum_count)

    def get_event_reporting_flush_interval(self):
        return self.__event_reporting_flush_interval
    event_reporting_flush_interval = property(get_event_reporting_flush_interval)

    def get_event_reporting_trap_exceptions(self):
        return self.__event_reporting_trap_exceptions
    event_reporting_trap_exceptions = property(get_event_reporting_trap_exceptions)
    
    def get_event_reporting_timestamp_granularity(self):
        return self.__event_reporting_timestamp_granularity
    event_reporting_timestamp_granularity = property(get_event_reporting_timestamp_granularity)

    def get_event_reporting_synchronous(self):
        return self.__event_reporting_synchronous
    event_reporting_synchronous = property(get_event_reporting_synchronous)
    
    def get_logging_verbose(self):
        return self.__logging_verbose
    logging_verbose = property(get_logging_verbose)

    def get_key_caching_unstructured(self):
        return self.__key_caching_unstructured
    key_caching_unstructured = property(get_key_caching_unstructured)
    
    def get_key_caching_structured(self):
        return self.__key_caching_structured
    key_caching_structured = property(get_key_caching_structured)
    
    def get_key_caching_encrypt(self):
        return self.__key_caching_encrypt
    key_caching_encrypt = property(get_key_caching_encrypt)
    
    def get_key_caching_ttl_seconds(self):
        return self.__key_caching_ttl_seconds
    key_caching_ttl_seconds = property(get_key_caching_ttl_seconds)

    def get_idp_provider(self):
        return self.__idp_provider
    idp_provider = property(get_idp_provider)
    
    def get_idp_ubiq_customer_id(self):
        return self.__idp_ubiq_customer_id
    idp_ubiq_customer_id = property(get_idp_ubiq_customer_id)
    
    def get_idp_token_endpoint_url(self):
        return self.__idp_token_endpoint_url
    idp_token_endpoint_url = property(get_idp_token_endpoint_url)
    
    def get_idp_tenant_id(self):
        return self.__idp_tenant_id
    idp_tenant_id = property(get_idp_tenant_id)
    
    def get_idp_client_secret(self):
        return self.__idp_client_secret
    idp_client_secret = property(get_idp_client_secret)
    
    def set(self):
        idp_all_set = False
        idp_in_use = self.__idp_provider != None
        if(idp_in_use):
            idp_all_set = (self.__idp_ubiq_customer_id != None
            and self.__idp_token_endpoint_url != None
            and self.__idp_tenant_id != None
            and self.__idp_client_secret != None)
            
        # If IDP provider is set, everything else needs set
        # Otherwise, continue as normal
        return (self.__event_reporting_wake_interval != None 
                and self.__event_reporting_minimum_count != None 
                and self.__event_reporting_flush_interval != None
                and self.__logging_verbose != None) and not (idp_all_set ^ idp_in_use)

class ubiqConfiguration(configInfo):#
    
    def load_config_dict(self, config_dict):
        if isinstance(config_dict, dict):
            if 'event_reporting' in config_dict:
                if 'wake_interval' in config_dict['event_reporting']:
                    self.__event_reporting_wake_interval = config_dict['event_reporting']['wake_interval']
                if 'minimum_count' in config_dict['event_reporting']:
                    self.__event_reporting_minimum_count = config_dict['event_reporting']['minimum_count']
                if 'flush_interval' in config_dict['event_reporting']:
                    self.__event_reporting_flush_interval = config_dict['event_reporting']['flush_interval']
                if 'trap_exceptions' in config_dict['event_reporting']:
                    self.__event_reporting_trap_exceptions = config_dict['event_reporting']['trap_exceptions']
                if 'timestamp_granularity' in config_dict['event_reporting']:
                    self.__event_reporting_timestamp_granularity = get_timestamp_granularity(config_dict['event_reporting']['timestamp_granularity'])
                if 'synchronous' in config_dict['event_reporting']:
                    self.__event_reporting_synchronous = config_dict['event_reporting']['synchronous']
            if 'logging' in config_dict:
                if 'verbose' in config_dict['logging']:
                    self.__logging_verbose = config_dict['logging']['verbose']
            if 'key_caching' in config_dict:
                if 'unstructured' in config_dict['key_caching']:
                    self.__key_caching_unstructured = config_dict['key_caching']['unstructured']
                if 'structured' in config_dict['key_caching']:
                    self.__key_caching_structured = config_dict['key_caching']['structured']
                if 'encrypt' in config_dict['key_caching']:
                    self.__key_caching_encrypt = config_dict['key_caching']['encrypt']
                if 'ttl_seconds' in config_dict['key_caching']:
                    self.__key_caching_ttl_seconds = config_dict['key_caching']['ttl_seconds']
            if 'idp' in config_dict:
                if 'provider' in config_dict['idp']:
                    self.__idp_provider = config_dict['idp']['provider']
                if 'ubiq_customer_id' in config_dict['idp']:
                    self.__idp_ubiq_customer_id = config_dict['idp']['ubiq_customer_id']
                if 'idp_token_endpoint_url' in config_dict['idp']:
                    self.__idp_token_endpoint_url = config_dict['idp']['idp_token_endpoint_url']
                if 'idp_tenant_id' in config_dict['idp']:
                    self.__idp_tenant_id = config_dict['idp']['idp_tenant_id']
                if 'idp_client_secret' in config_dict['idp']:
                    self.__idp_client_secret = config_dict['idp']['idp_client_secret']

    def load_config_file(self, config_file):
        try:
            with open(config_file) as json_file:
                config = json.load(json_file)
                self.load_config_dict(config)
        except FileNotFoundError:
            # If file doesn't exist, use defaults
            pass

    def set_defaults(self):
        self.__event_reporting_wake_interval = 10
        self.__event_reporting_minimum_count = 50
        self.__event_reporting_flush_interval = 90
        self.__event_reporting_trap_exceptions = False
        self.__event_reporting_timestamp_granularity = TimestampGranularity.MICROS
        self.__event_reporting_synchronous = False
        self.__logging_verbose = False
        self.__key_caching_unstructured = True
        self.__key_caching_structured = True
        self.__key_caching_encrypt = False
        self.__key_caching_ttl_seconds = 1800
        self.__idp_provider = None
        self.__idp_ubiq_customer_id = None
        self.__idp_token_endpoint_url = None
        self.__idp_tenant_id = None
        self.__idp_client_secret = None

    def __init__(self, config_file = None, config_dict = None):
        self.__event_reporting_wake_interval = None
        self.__event_reporting_minimum_count = None
        self.__event_reporting_flush_interval = None
        self.__event_reporting_trap_exceptions = None
        self.__event_reporting_timestamp_granularity = None
        self.__event_reporting_synchronous = None
        self.__logging_verbose = None
        self.__key_caching_unstructured = None
        self.__key_caching_structured = None
        self.__key_caching_encrypt = None
        self.__key_caching_ttl_seconds = None
        self.__idp_provider = None
        self.__idp_ubiq_customer_id = None
        self.__idp_token_endpoint_url = None
        self.__idp_tenant_id = None
        self.__idp_client_secret = None

        self.set_defaults()
        
        if (config_file == None):
            from os.path import expanduser
            home = expanduser("~")
            config_file = os.path.join(home, ".ubiq", "configuration")

        if os.path.exists(config_file):
            self.load_config_file(config_file)

        # Merge config dict onto Config File (if exists)
        if (config_dict != None):
            self.load_config_dict(config_dict)

        configInfo.__init__(
            self, 
            self.__event_reporting_wake_interval, 
            self.__event_reporting_minimum_count,
            self.__event_reporting_flush_interval,
            self.__event_reporting_trap_exceptions,
            self.__event_reporting_timestamp_granularity,
            self.__event_reporting_synchronous,
            self.__logging_verbose,
            self.__key_caching_unstructured,
            self.__key_caching_structured,
            self.__key_caching_encrypt,
            self.__key_caching_ttl_seconds,
            self.__idp_provider,
            self.__idp_ubiq_customer_id,
            self.__idp_token_endpoint_url,
            self.__idp_tenant_id,
            self.__idp_client_secret)
        
        # If verbose, warn user if M2Crypto will not be used.
        if self.__logging_verbose:
            # Package name is case-sensitive.
            if importlib.util.find_spec('M2Crypto') is None:
                print('M2Crypto not found on system. Defaulting to Cryptography (slower, but still functional).')



#!/usr/bin/env python3

import atexit
import base64
import configparser
from datetime import datetime, timedelta, timezone
import json
import os
import http
import urllib.error
import requests
import urllib
from . import UBIQ_HOST
from .configuration import ubiqConfiguration
from .events import events, eventsProcessor, syncEventsProcessor

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend

class credentialsInfo:

    def __init__(self, access_key_id, secret_signing_key, secret_crypto_access_key, host, config_file, library_label, config_obj, idp_username, idp_password):
        self.__access_key_id = access_key_id
        self.__secret_signing_key = secret_signing_key
        self.__secret_crypto_access_key = secret_crypto_access_key
        
        self.__idp_username = idp_username
        self.__idp_password = idp_password
        self.__initialized = False
        
        self.__idp_cert_base64 = None
        self.__idp_cert_expires = None
        self.__encrypted_private_key = None
        
        self.__host = host
        if (not self.__host.lower().startswith('http')):
            self.__host = "https://" + self.__host

        if (config_file != None and config_obj != None):
               raise RuntimeError("Either config_file or config_obj should be set, but not both.")

        if (config_obj != None):
            self.__configuration = config_obj
        else:
            self.__configuration = ubiqConfiguration(config_file)
        
        # Event Tracking
        self.__events = events(self, self.__configuration, library_label)
        if self.__configuration.get_event_reporting_synchronous():
            self.__eventsProcessor = syncEventsProcessor(self.__configuration, self.__events)
        else:
            self.__eventsProcessor = eventsProcessor(self.__configuration, self.__events)
            self.__eventsProcessor.start()

    def get_access_key_id(self):
        return self.__access_key_id
    access_key_id=property(get_access_key_id)

    def get_secret_signing_key(self):
        return self.__secret_signing_key
    secret_signing_key=property(get_secret_signing_key)

    def get_secret_crypto_access_key(self):
        return self.__secret_crypto_access_key
    secret_crypto_access_key = property(get_secret_crypto_access_key)

    def get_host(self):
        return self.__host
    host = property(get_host)

    def get_configuration(self):
        return self.__configuration
    configuration = property(get_configuration)
    
    def get_encrypted_private_key(self):
        return self.__encrypted_private_key
    encrypted_private_key = property(get_encrypted_private_key)
    
    def get_idp_cert_base64(self):
        return self.__idp_cert_base64
    idp_cert_base64 = property(get_idp_cert_base64)

    def set(self):
        return (
                # Traditional
                self.__access_key_id != None and 
                self.__secret_signing_key != None and 
                self.__secret_crypto_access_key != None and 
                self.__access_key_id.strip() != "" and 
                self.__secret_signing_key.strip() != "" and 
                self.__secret_crypto_access_key.strip() != ""
            ) or (
                # IDP
                self.__idp_password != None and 
                self.__idp_password != None
            )
            
    
    def add_reporting_user_defined_metadata(self, data):
        self.__events.add_user_defined_metadata(data)
    
    def get_copy_of_usage(self):
        return self.__events.list_events()
    
    def get_event_count(self):
        return self.__events.get_events_count()

    # Forward events to event queue
    def add_event(self, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count):
        return self.__events.add_event(self.__access_key_id, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count)
    
    # Manually trigger process events
    def process_events(self):
        return self.__eventsProcessor.process()
    
    ## IDP RELEVANT FUNCTIONS ##
    
    def get_oauth_token(self):
        config = self.__configuration
        provider = config.get_idp_provider()
        endpoint = config.get_idp_token_endpoint_url()

        query = {
            'client_id': config.get_idp_tenant_id(),
            'client_secret': config.get_idp_client_secret(),
            'username': self.__idp_username,
            'password': self.__idp_password,
            'grant_type': 'password'
        }
        if provider == 'okta':
            query['scope'] = 'openid offline_access okta.users.read okta.groups.read'
        elif provider == 'entra':
            query['scope'] = f'api://{config.get_idp_tenant_id()}/.default'
        else:
            raise RuntimeError(f'Unknown or no IDP provider specified: {provider} Check your configuration and try again.')
        
        resp = requests.post(
            endpoint,
            data=query,
            headers={
                'Accept': 'application/json',
                'Cache-control': 'no-cache',
                'Content-type': 'application/x-www-form-urlencoded',
            })

        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                endpoint, resp.status_code, f'Status ({resp.status_code}) Unable to fetch token from {endpoint}',
                resp.headers, resp.content
            )
        else:
            return json.loads(resp.content.decode())
    
    def get_idp_token_and_cert(self):
        self.token = self.get_oauth_token()
        self.sso = self.get_sso(self.token['access_token'], self.csr)
        api_cert = x509.load_pem_x509_certificate(self.sso['api_cert'].encode('utf-8'))
        self.__idp_cert_base64 = base64.b64encode(self.sso['api_cert'].encode('utf-8')).decode('utf-8')
        self.__idp_cert_expires = api_cert.not_valid_after_utc - timedelta(minutes=1)

    def renew_idp_cert(self):
        if self.is_idp():
            if self.__idp_cert_expires < datetime.now(timezone.utc):
                self.get_idp_token_and_cert()
    
    # Required call if IDP mode
    # Sets up the handshake between IDP & Ubiq
    def init(self):       
        srsa_bytes = base64.b64encode(os.urandom(33))
        self.__secret_crypto_access_key = srsa_bytes.decode('utf-8')        
        key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=4096
        )
        
        private_key = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(srsa_bytes)
        )
        
        csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, base64.b64encode(os.urandom(18)).decode('utf-8')),
            x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'California'),
            x509.NameAttribute(NameOID.LOCALITY_NAME, 'San Diego'),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'Ubiq Security, Inc.'),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, 'Ubiq Platform')
        ])).sign(key, hashes.SHA256())
        
        self.csr = csr.public_bytes(serialization.Encoding.PEM)
        
        self.get_idp_token_and_cert()
        
        self.__access_key_id = self.sso['public_value']
        self.__secret_signing_key = self.sso['signing_value']

        self.__encrypted_private_key = private_key.decode('utf-8')
        self.__initialized = True

    def get_sso(self, access_token, csr):
        url = f'{self.host}/{self.__configuration.get_idp_ubiq_customer_id()}/api/v3/scim/sso'
        resp = requests.post(
            url=url,
            data=json.dumps({'csr': csr.decode('utf-8')}),
            headers={
                'Authorization': f'Bearer {access_token}',
                'Accept': 'application/json',
                'Cache-control': 'no-cache',
                'content-type': 'application/json'
            }
        )
        
        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(url, resp.status_code, f'Status {resp.status_code} Unable to validate token {url}')
        else:
            return json.loads(resp.content.decode())

    # Easy method for checking if running in IDP mode
    def is_idp(self):
        ret = self.__idp_username != None and len(self.__idp_username) > 0
        
        if (ret and not self.__initialized):
            raise RuntimeError(f'Credentials.init() has not been called or failed but is required when using IDP authentication')
        
        return ret
        
    ## END IDP RELEVANT FUNCTIONS ##

# TODO: Rename configCredentials and load_config_file cause they are confusing. (configurableCredentials? load_credentials_file?)
class configCredentials(credentialsInfo):

    def load_config_file(self, credentials_file, profile):
        config = configparser.ConfigParser(interpolation=None)
        config.read(credentials_file)

        # Create empty dictionaries for the default and supplied profile
        d = {}
        p = {}

        # get the default profile if there is one
        if (config.has_section('default')):
            d = config['default']

        # get the supplied profile if there is one
        if (config.has_section(profile)):
            p = config[profile]

        # Use given profile if it is available, otherwise use default.
        self.__access_key_id= p.get('access_key_id', d.get('access_key_id'))
        self.__secret_signing_key = p.get('secret_signing_key', d.get('secret_signing_key'))
        self.__secret_crypto_access_key = p.get('secret_crypto_access_key', d.get('secret_crypto_access_key'))
        self.__host = p.get('SERVER', d.get('SERVER', UBIQ_HOST))
        self.__idp_username = p.get('IDP_USERNAME', d.get('IDP_USERNAME'))
        self.__idp_password = p.get('IDP_PASSWORD', d.get('IDP_PASSWORD'))


    def __init__(self, credentials_file = None, profile = "default", config_file = None, library_label = None, config_obj = None,):
        self.__access_key_id = None
        self.__secret_signing_key = None
        self.__secret_crypto_access_key = None
        self.__host = None
        self.__idp_username = None
        self.__idp_password = None

        if (credentials_file == None):
            from os.path import expanduser
            home = expanduser("~")
            credentials_file = os.path.join(home, ".ubiq", "credentials")

        if os.path.exists(credentials_file):
            self.load_config_file(credentials_file, profile)

        credentialsInfo.__init__(self, self.__access_key_id , self.__secret_signing_key, self.__secret_crypto_access_key, self.__host, config_file, library_label, config_obj, self.__idp_username, self.__idp_password)

        if (not self.set()):
            if (self.__access_key_id == None or self.__access_key_id.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'access_key_id' value in profile '{1}'.".format(credentials_file, profile))
            elif (self.__secret_signing_key == None or self.__secret_signing_key.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'secret_signing_key' value in profile '{1}'.".format(credentials_file, profile))
            elif(self.__secret_crypto_access_key == None or self.__secret_crypto_access_key.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'secret_crypto_access_key' value in profile '{1}'.".format(credentials_file, profile))


class credentials(credentialsInfo):

    def __init__(self, access_key_id = None, secret_signing_key = None, secret_crypto_access_key = None, host = UBIQ_HOST, config_file = None, library_label = None, config_obj = None, idp_username = None, idp_password = None):
        # If supplied value is None, use ENV variable, otherwise use supplied value.
        # If env value isn't set, use the supplied value anyways (None) but prevent an exception
        self.__access_key_id = (access_key_id, os.getenv('UBIQ_ACCESS_KEY_ID', access_key_id)) [access_key_id == None]
        self.__secret_signing_key = (secret_signing_key, os.getenv('UBIQ_SECRET_SIGNING_KEY', secret_signing_key)) [secret_signing_key == None]
        self.__secret_crypto_access_key = (secret_crypto_access_key, os.getenv('UBIQ_SECRET_CRYPTO_ACCESS_KEY', secret_crypto_access_key)) [secret_crypto_access_key == None]
        config_file = (config_file, os.getenv('UBIQ_CONFIGURATION_FILE_PATH', config_file)) [config_file == None]
        self.__host = (host, os.getenv('UBIQ_SERVER', host)) [host == None]
        self.__idp_username = (idp_username, os.getenv('IDP_USERNAME', idp_username))[idp_username == None]
        self.__idp_password = (idp_password, os.getenv('IDP_PASSWORD', idp_password))[idp_password == None]

        credentialsInfo.__init__(self, self.__access_key_id,
                                 self.__secret_signing_key,
                                 self.__secret_crypto_access_key, 
                                 self.__host,
                                 config_file,
                                 library_label,
                                 config_obj,
                                 self.__idp_username,
                                 self.__idp_password)
        
        if (not self.set()):
            if (self.__access_key_id == None or self.__access_key_id.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_ACCESS_KEY_ID' not set.")
            elif (self.__secret_signing_key == None or self.__secret_signing_key.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_SECRET_SIGNING_KEY' not set.")
            elif(self.__secret_crypto_access_key == None or self.__secret_crypto_access_key.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_SECRET_CRYPTO_ACCESS_KEY' not set.")



#!/usr/bin/env python3

import base64
import http
import json
import requests
import urllib
import time
import copy

from ..auth import http_auth
from .lib import ffx


import cryptography.hazmat.primitives as crypto
import cryptography.hazmat.primitives.serialization as serialize 
from cryptography.hazmat.backends import default_backend as crypto_backend

def strConvertRadix(s, ics, ocs):
    return ffx.NumberToString(len(ocs), ocs,
                              ffx.StringToNumber(len(ics), ics, s),
                              len(s))

def fmtInput(s, pth, ics, ocs, rules = []):
    fmt = ''
    trm = '%s'%(s)
    
    # Check if there's a passthrough rule. If not, create for legacy passthrough.
    if not any(rule.get('type') == 'passthrough' for rule in rules):
        rules.insert(0, {'type': 'passthrough', 'value': pth, 'priority': 1})
        
    # Sort the rules by priority
    rules.sort(key=lambda x: x['priority'])
    for idx, rule in enumerate(rules):
        if(rule['type'] == 'passthrough'):
            pth = rule['value']
            o = ''
            for c in trm:
                if c in pth:
                    fmt += c
                else:
                    fmt += ocs[0]
                    o += c
            trm = o
        elif(rule['type'] == 'prefix'):
            rules[idx]['buffer'] = trm[:rule['value']]
            trm = trm[rule['value']:]
        elif(rule['type'] == 'suffix'):
            rules[idx]['buffer'] = trm[(-1 * rule['value']):]
            trm = trm[:(-1 * rule['value'])]
        else:
            raise RuntimeError('Ubiq Python Library does not support rule type "%s" at this time.'%(rule['type']))

    # Validate final string contains only allowed characters.
    if not all((c in ics) for c in trm):
        raise RuntimeError('Invalid input string character(s)')

    return fmt, trm, rules

def encKeyNumber(s, ocs, n, sft):    
    return ocs[ocs.find(s[0]) + (int(n) << sft)] + s[1:]

def decKeyNumber(s, ocs, sft):
    charBuf = s[0]
    encoded_value = ocs.find(charBuf)
    key_num = encoded_value >> sft

    return ocs[encoded_value - (key_num << sft)] + s[1:], key_num

def fmtOutput(fmt, s, pth, rules):
    # Sort the rules by decreasing priority
    rules.sort(key=lambda x: x['priority'], reverse=True)

    for rule in rules:
        if(rule['type'] == 'passthrough'):
            o = ''
            for c in fmt:
                if c not in pth:
                    o, s = o + s[0], s[1:]
                else:
                    o += c
                
            if len(s) > 0:
                raise RuntimeError('mismatched format and output strings')
            s = o
        elif(rule['type'] == 'prefix'):
            s = rule['buffer'] + s
        elif(rule['type'] == 'suffix'):
            s = s + rule['buffer']
        else:
            raise RuntimeError('Ubiq Python Library does not support rule type "%s" at this time.'%(rule['type']))

    return s

def fetchDataset(creds, dataset_name):
    papi = creds.access_key_id
    sapi = creds.secret_signing_key
    host = creds.host
    
    config = creds.configuration
    ttl_seconds = config.key_caching_ttl_seconds
    unstructured_cache_enabled = config.key_caching_structured
    
    if (not unstructured_cache_enabled or
        not papi in fetchDataset.cache or
        not dataset_name in fetchDataset.cache[papi] or
        fetchDataset.cache[papi][dataset_name]["expires"] < time.time()):
        
        if config.logging_verbose:
            print('****** PERFORMING EXPENSIVE CALL ----- fetchDataset')
        
        url = host + '/api/v0/ffs'
        url += '?ffs_name=' + dataset_name
        url += '&papi=' + papi
        
        resp = requests.get(
            url, 
            auth=http_auth(papi, sapi))
        
        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                url, resp.status_code,
                http.HTTPStatus(resp.status_code).phrase,
                resp.headers, resp.content)
        if not papi in fetchDataset.cache:
            fetchDataset.cache[papi] = {}
        fetchDataset.cache[papi][dataset_name] = { "dataset" : json.loads(resp.content.decode()), "expires" : time.time() + ttl_seconds }

    return fetchDataset.cache[papi][dataset_name]["dataset"]
fetchDataset.cache = {}

def flushDataset(papi = None, dataset_name = None):
    if papi == None:
        fetchDataset.cache = {}
    elif papi in fetchDataset.cache:
        if dataset_name == None:
            del fetchDataset.cache[papi]
        elif dataset_name in fetchDataset.cache[papi]:
            del fetchDataset.cache[papi][dataset_name]
            
def add_to_fetchkey_cache(papi, dataset_name, n, key, ttl_seconds):
    cache_entry = { "key" : key, "expires": time.time() + ttl_seconds }
    
    if not papi in fetchKey.cache:
        fetchKey.cache[papi] = {}
    if not dataset_name in fetchKey.cache[papi]:
        fetchKey.cache[papi][dataset_name] = {}

    # the -1 entry points to the "current" key at the
    # server. it is cached so that the next caller that
    # wants the "current" key can get it, but it should
    # be timed-out occasionally in case the "current"
    # pointer changes at the server.
    #
    # that timeout is future work
    if n == -1:
        # -1 can be an index because keys are stored
        # in a dictionary, not a list
        fetchKey.cache[papi][dataset_name][n] = cache_entry

    # also cache the key at its "real" identifier
    n = int(key['key_number'])
    fetchKey.cache[papi][dataset_name][n] = cache_entry

def fetchKey(creds, dataset_name, n = -1):
    papi = creds.access_key_id
    sapi = creds.secret_signing_key
    srsa = creds.secret_crypto_access_key
    host = creds.host
    
    config = creds.configuration
    ttl_seconds = config.key_caching_ttl_seconds
    structured_cache_enabled = creds.configuration.key_caching_structured
    cache_encrypted = creds.configuration.key_caching_encrypt
    
    key = None
    
    if (not structured_cache_enabled or
        not papi in fetchKey.cache or
        not dataset_name in fetchKey.cache[papi] or
        not n in fetchKey.cache[papi][dataset_name] or
        fetchKey.cache[papi][dataset_name][n]["expires"] < time.time()):
        
        if config.logging_verbose:
            print('****** PERFORMING EXPENSIVE CALL ----- fetchKey')

        url = host + '/api/v0/fpe/key'
        url += '?ffs_name=' + dataset_name
        url += '&papi=' + papi
        if n >= 0:
            url += '&key_number=' + str(n)
        query = {}
        # IDP mode requires passing the idp_cert to the server
        if creds.is_idp():
            creds.renew_idp_cert()
            # Query param since we're doing a GET
            url += '&payload_cert=' + urllib.parse.quote_plus(creds.idp_cert_base64)
        resp = requests.get(
            url, 
            auth=http_auth(papi, sapi))
        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                url, resp.status_code,
                http.HTTPStatus(resp.status_code).phrase,
                resp.headers, resp.content)
        key = json.loads(resp.content.decode())
        
        # IDP mode has a local private key, need to override that key since nothing will be returned from server
        if creds.is_idp():
            key['encrypted_private_key'] = creds.encrypted_private_key
        
        if structured_cache_enabled and cache_encrypted:
            add_to_fetchkey_cache(papi, dataset_name, n, copy.deepcopy(key), ttl_seconds)
    else:
        key = fetchKey.cache[papi][dataset_name][n]["key"]
    
    if key != None and not 'unwrapped_data_key' in key:
        prvkey = serialize.load_pem_private_key(
            key['encrypted_private_key'].encode(), srsa.encode(),
            crypto_backend())

        key['unwrapped_data_key'] = prvkey.decrypt(
            base64.b64decode(key['wrapped_data_key']),
            crypto.asymmetric.padding.OAEP(
                mgf=crypto.asymmetric.padding.MGF1(
                    algorithm=crypto.hashes.SHA1()),
                algorithm=crypto.hashes.SHA1(),
                label=None))

        if structured_cache_enabled and not cache_encrypted:
            add_to_fetchkey_cache(papi, dataset_name, n, copy.deepcopy(key), ttl_seconds)
    return key
fetchKey.cache = {}

def allKeysToNInCache(papi, dataset_name, n):
    present = True
    for i in range(0,n+1):
        present = present and (i in fetchKey.cache[papi][dataset_name])
    return present

def fetchAllKeys(creds, dataset_name):
    papi = creds.access_key_id
    sapi = creds.secret_signing_key
    srsa = creds.secret_crypto_access_key
    host = creds.host
    
    config = creds.configuration
    ttl_seconds = config.key_caching_ttl_seconds
    structured_cache_enabled = config.key_caching_structured
    cache_encrypted = config.key_caching_encrypt
    
    if config.logging_verbose:
        print('****** PERFORMING EXPENSIVE CALL ----- fetchAllKeys')
    
    url=f"{host}/api/v0/fpe/def_keys?ffs_name={dataset_name}&papi={papi}"
    
    if creds.is_idp():
        creds.renew_idp_cert()
        # Query param since we're doing a GET
        url += '&payload_cert=' + urllib.parse.quote_plus(creds.idp_cert_base64)
    
    resp = requests.get(url, auth=http_auth(papi, sapi))
    
    all_keys = {}

    if resp.status_code != http.HTTPStatus.OK:
        raise urllib.error.HTTPError(
            url, resp.status_code,
            http.HTTPStatus(resp.status_code).phrase,
            resp.headers, resp.content)
    keys = json.loads(resp.content.decode())
    
    # IDP mode has a local private key, need to override that key since nothing will be returned from server
    if creds.is_idp():
        keys[dataset_name]['encrypted_private_key'] = creds.encrypted_private_key

    if structured_cache_enabled:
        if not papi in fetchKey.cache:
            fetchKey.cache[papi] = {}
        if not dataset_name in fetchKey.cache[papi]:
            fetchKey.cache[papi][dataset_name] = {}

    prvkey = serialize.load_pem_private_key(
        keys[dataset_name]['encrypted_private_key'].encode(), srsa.encode(),
        crypto_backend())
    
    for i, enc_key in enumerate(keys[dataset_name]['keys']):
        if (structured_cache_enabled and # Cache is Enabled
            i in fetchKey.cache[papi][dataset_name] and # Key in cache
            fetchKey.cache[papi][dataset_name][i]["expires"] > time.time()): # Cache isnt expired
            key = fetchKey.cache[papi][dataset_name][i]["key"]
            
            if 'unwrapped_data_key' in key:
                all_keys[i] = key
                continue

        key = {
            'encrypted_private_key': keys[dataset_name]['encrypted_private_key'],
            'wrapped_data_key': enc_key,
            'key_number': i
        }
        
        # Store cache encrpypted (don't store unwrapped)
        if structured_cache_enabled and cache_encrypted:
            cache_entry = { "key" : key, "expires": time.time() + ttl_seconds }
            fetchKey.cache[papi][dataset_name][i] = cache_entry
        
        key['unwrapped_data_key'] = prvkey.decrypt(
            base64.b64decode(key['wrapped_data_key']),
            crypto.asymmetric.padding.OAEP(
                mgf=crypto.asymmetric.padding.MGF1(
                    algorithm=crypto.hashes.SHA1()),
                algorithm=crypto.hashes.SHA1(),
                label=None))
        
        # Store cache unencrypted
        if structured_cache_enabled and not cache_encrypted: 
            cache_entry = { "key" : key, "expires": time.time() + ttl_seconds }
            fetchKey.cache[papi][dataset_name][i] = cache_entry

        all_keys[i] = key

    return all_keys

def fetchCurrentKeys(creds, dataset_name):
    return fetchAllKeys(creds, dataset_name)


def flushKey(papi = None, dataset_name = None, n = None):
    if papi == None:
        fetchKey.cache = {}
    elif papi in fetchKey.cache:
        if dataset_name == None:
            del fetchKey.cache[papi]
        elif dataset_name in fetchKey.cache[papi]:
            if n == None:
                del fetchKey.cache[papi][dataset_name]
            elif n in fetchKey.cache[papi][dataset_name]:
                del fetchKey.cache[papi][dataset_name][n]
